/*=============================================================================
 * 
 * Robot Framework 4
 * Browser object
 *
 * (c) 2015 Statistics Netherlands
 *
 * Revision history:
 * 02 march    2015: EHVL: split off from framework.js
 * 01 april    2015: EHVL: added "--ignore-ssl-errors=true" to phantomjs'
 *                         list of cli arguments
 * 14 april    2015: EHVL: removed "--ignore-ssl-errors=true" 
 *                         (no longer necessary as SSL certs up-to-date now)
 * 12 mei      2015: EHVL: ignore links that no longer exist
 *    januari  2016: EHVL: complete rewrite for better error logging
 * 03 februari 2016: EHVL: added error message when stopping browser
 *=============================================================================
 */

var webdriver = require("selenium-webdriver");
var phantom = require("selenium-webdriver/phantomjs");
var _ = require("underscore");
var cheerio = require("cheerio");

var tech = require("./levellogger");
var user = require("./logger");
var opcodes = require("./opcodes");

var wdErrCode = webdriver.error.ErrorCode;

/* Timeout for page loads, in milliseconds. We set this to 1 minute.
 * Default behaviour is to wait endlessly, which is not very
 * convenient, as it can cause the framework to hang. */
const PAGELOAD_TIMEOUT_MS = 1 * 60 * 1000;

/* Timeout for wait functions, in milliseconds */
const WAIT_TIMEOUT_MS = 2 * 60 * 1000;

/* The number of browser actions after which the browser is restarted */
const N_BROWSER_ACTIONS_THRESHOLD = 100;

/* The default setting whether to restart the browser after N_BROWSER_ACTIONS_THRESHOLD
 * browser actions. */
const IS_BROWSER_RESTART_DEFAULT = true;

/* The number of times that recover() tries to recover the browser state 
 * Note: recover will always try to recover at least once. Therefore, a value of
 * N_RECOVER_TRIES smaller than 1 has the same effect as a value of 1. */
const N_RECOVER_TRIES = 3;

/* The number of times a page is reloaded before quitting due to 
 * connection failure. */
const N_CONNECTION_INTERRUPTED_TRIES = 10;

/* The delay (in milliseconds) between reload attempts */
const RELOAD_DELAY_MS = 30 * 1000;

/*------------------------------------------------------------------------------
 * findWebDriverURL()
 *
 * Finds the webdriver host URL and port number.
 * Note that this is a rather ugly hack: we deliberately pass an invalid XPATH
 * expression to the web driver, which returns an error; inside the error 
 * message we can find the server URL, as part of the JSON wire request 
 * sent to the browser.
 *
 * If there is a more direct way to obtain the server URL I am all ears.
 * Inspection of the selenium code has shown that it is a private member
 * of some deep-down object called HttpClient.
 *------------------------------------------------------------------------------
 */
function findWebDriverURL(
    driver        // selenium webdriver object
) {

    return driver.findElement({xpath:"1"}).then(function(elt) {
        // This should never happen. But we prepare for the worst...
        tech.fatal("Internal error: unable to determine webdriver host URL");
    }, function(err) {
        if (err instanceof webdriver.error.WebDriverError 
            && err.code == wdErrCode.INVALID_SELECTOR_ERROR) {
            
            var serverUrl = 
                "http://" + JSON.parse(err.message).request.headers.Host;

            return serverUrl;
        } else {
            throw err;
        }
    }); 
}

/*------------------------------------------------------------------------------
 * Browser constructor
 *
 * Note that the selenium driver is no longer initialized here. Instead, 
 * Browser.start() must be called separately. The reason for this is that
 * the code must make sure that the phantomjs browser runs correctly before
 * opening the first page.
 *------------------------------------------------------------------------------
 */
function Browser(
    userAgent, // (optional) user agent to use. If not set, the default 
               // phantomjs user agent will be used.
    logName,   // File name of the browser log
    logLevel,  // Log level of the browser log
    isBrowserRestart // Should the browser be restarted periodically?
) {
    this._nEmpty = 0;
    var self = this;

    // Set up browser capabilities
    this._browserCapabilities = webdriver.Capabilities.phantomjs();

    // Don't load images
    // Edit (20151111) Do load images because of memory leak in phantomjs 2.0.0
    // Edit (20160105) Revert to phantomjs 1.9.8; don't load images
    // Edit (20160304) EHVL: this no longer works with selenium 2.52. Use CLI argument instead
    // Edit (20160304) EHVL: looks like false does not work, while "false" does.
    //this._browserCapabilities.set("phantomjs.page.settings.loadImages", "false");

    // Use the given user agent, if set
    if (userAgent) {
        tech.message("INFO", "Using user agent: " + userAgent);
        this._browserCapabilities.set(
            "phantomjs.page.settings.userAgent", userAgent);
    } else {
        tech.message("INFO", "Using default phantomjs user agent");
    }
    
    // Browser log file setup
    if (logLevel) {
        this._browserCapabilities.set("phantomjs.cli.args", 
            ["--webdriver-logfile=" + logName, 
             "--webdriver-loglevel=" + logLevel, 
             "--load-images=false"]);
    } else {
        // Don't generate a browser log file
        // Note that the space after none is significant
        this._browserCapabilities.set("phantomjs.cli.args", 
            ["--webdriver-logfile=none ",
             "--load-images=false"]);
    }
    
    this._url = null;
    this._history = [];
        
    // Determine whether to periodically restart the browser
    if (typeof(isBrowserRestart) === "undefined") {
        tech.info("isBrowserRestart not specified. Using default: " + IS_BROWSER_RESTART_DEFAULT)
        this.isBrowserRestart = IS_BROWSER_RESTART_DEFAULT;
    } else {
        tech.info("Using specified value for isBrowserRestart: " + isBrowserRestart)
        this.isBrowserRestart = isBrowserRestart;
    }
        
    // Default wait callbacks
    this._waitCallbacks = {
        sleep: function(args) {
            var duration = args.duration;
            if (typeof duration !== "number") {
                tech.fatal("No valid 'duration' argument for " + 
                           "wait operation 'sleep'");
            }
            return this._driver.sleep(duration);
        },
        waitTillVisible: function(args) {
            var selector;
            if (typeof args.xpathSelector === "string" && args.xpathSelector) {
                selector = { xpath: args.xpathSelector };
            } else if (typeof args.cssSelector === "string" && args.cssSelector) {
                selector = { css: args.cssSelector };
            } else {
                tech.fatal("No valid xpath or css selector for " + 
                           "wait operation 'waitTillVisible'");
            }
            var self = this;
            return this._driver.wait(function() {
                return self.findElementsAtomic(selector).then(function(elements) {
                    return elements.length > 0;
                });
            }, WAIT_TIMEOUT_MS);
        },            
        waitTillInvisible: function(args) {
            var selector;
            if (typeof args.xpathSelector === "string" && args.xpathSelector) {
                selector = { xpath: args.xpathSelector };
            } else if (typeof args.cssSelector === "string" && args.cssSelector) {
                selector = { css: args.cssSelector };
            } else {
                tech.fatal("No valid xpath or css selector for " + 
                           "wait operation 'waitTillInvisible'");
            }
            var self = this;
            return this._driver.wait(function() {
                return self.findElementsAtomic(selector).then(function(elements) {
                    return elements.length == 0;
                });
            }, WAIT_TIMEOUT_MS);
        }
    };
        
    // Default check callbacks
    this._checkCallbacks = {
        checkSucceed: function() {
            return webdriver.promise.fulfilled(true);
        },
        checkFail: function() {
            return webdriver.promise.fulfilled(false);
        }
    };
}

/*------------------------------------------------------------------------------
 * Browser.start()
 *
 * Starts phantomjs browser with capabilities as set by the constructor
 *------------------------------------------------------------------------------
 */
Browser.prototype.start = function() {
    
    tech.debug("Starting browser");

    var self = this;
    
    try {
        this._driver = new phantom.Driver(this._browserCapabilities);
    }
    catch(err) {
        tech.error("Error while starting browser: " + err.message);
        self.stop();
        return Browser.prototype.start.call(self);
    }
    
    this._nBrowserActions = 0;
    this._nBrowserErrors = 0;

    return webdriver.promise.all([
        findWebDriverURL(this._driver).then(function(url) {
            tech.info("Webdriver URL: " + url);
        }),
        this._driver.getSession().then(function(session) {
            tech.info("Webdriver session ID: " + session.getId());
        }),
        this._driver.manage().window().setSize(1280, 1024),
        this._driver.manage().timeouts().pageLoadTimeout(PAGELOAD_TIMEOUT_MS)
    ]).thenCatch(function(err) {
        //Try again
        tech.error("Unable to initialize browser: " + err.message);
        self.stop();
        return Browser.prototype.start.call(self);
    });
}

/*------------------------------------------------------------------------------
 * Browser.stop()
 *
 * Stops the browser
 *------------------------------------------------------------------------------
 */
Browser.prototype.stop = function() {
    
    tech.debug("stopping browser");
    if (this._driver) {
        return this._driver.quit().thenCatch(function(err) {
            // Ignore all errors, but log them anyway
            tech.error("Error while stopping the browser, continuing regardless: " + err.message)
        });
    }
}

/*------------------------------------------------------------------------------
 * Browser.restart()
 *
 * Restarts the browser, with the same capabilities as before
 *------------------------------------------------------------------------------
 */
Browser.prototype.restart = function() {
    
    var self = this;
    
    return self.stop().then(function() {
        return self.start();
    });
}

/*------------------------------------------------------------------------------
 * Browser.recover()
 *
 * Restarts the browser and recovers the browser to the state before
 * the browser action that led to the crash
 *------------------------------------------------------------------------------
 */
Browser.prototype.recover = function(nTries) {
    
    var self = this;
    
    if (typeof(nTries) === "undefined") nTries = 0;
    
    return self._driver.getSession().then(function () {
        tech.debug("Session found. Restarting browser");
        self.restart(); 
    }, 
    function() { 
        tech.debug("No session found. Starting browser");
        self.start(); //driver does not exist; just start
    }).then(function() {
    
        tech.debug("Recovering browser state");
        
        // make a copy of the self._history array
        var history = self._history.slice();
        
        /* Remove the last element from the history, because we want
         * to recover to the state before the last browser action.
         * We do this with a local copy because the recovery itself
         * could fail, in which case we need self._history intact. */ 
        if (history.length > 0) {
            history.pop();
        }

        return webdriver.promise.map(history, function(historyItem) {
            return self.execute(historyItem.instruction, historyItem.wait, historyItem.check, true);
        }).then(function() {
            tech.debug("Recovery complete");
            
            /* Now that the state is succesfully recovered we can remove the
             * last element from the "master" history in self._history. */
            if (self._history.length > 0) {
                self._history.pop();
            }
        });
    }).catch(function(err) {
        if (err instanceof Error 
            && (err.message.indexOf("ECONNREFUSED") != -1 
                || err.message.indexOf("Server terminated early") != -1)) {

            tech.error("Browser crashed during recovery");
            nTries += 1;
            if (nTries < N_RECOVER_TRIES) {
                return Browser.prototype.recover.call(self, nTries);
            } else {
                tech.fatal("Unable to recover after " 
                           + N_RECOVER_TRIES + " tries. Quitting.");
            }
        } else if (err instanceof Error
                   && err.message == "Internet connection failed") {
            // Internet connection failure dealt with the same as 
            // browser crash, but counted differently.
            tech.error("Connection failure during recovery");
            return self._driver.sleep(RELOAD_DELAY_MS).then(function() {
                return Browser.prototype.recover.call(self, nTries);
            });
        } else {
            tech.fatal("Uncaught error during recovery. Quitting: " + err.message);
        }
    });
}                                

/*------------------------------------------------------------------------------
 * Browser.registerWaitCallbacks()
 *
 * Add custom wait callbacks to the browser object
 *------------------------------------------------------------------------------
 */
Browser.prototype.registerWaitCallbacks = function(
    newCallbacks  // custom wait callbacks to add
) {

    _.extend(this._waitCallbacks, newCallbacks);
}

/*------------------------------------------------------------------------------
 * Browser.registerCheckCallbacks()
 *
 * Add custom check callbacks to the browser object
 *------------------------------------------------------------------------------
 */
Browser.prototype.registerCheckCallbacks = function(
    newCallbacks  // custom check callbacks to add
) {

    _.extend(this._checkCallbacks, newCallbacks);
}

/*------------------------------------------------------------------------------
 * Browser.execute()
 *
 * Executes an instruction (opcode + arguments) in the browser and wait for the
 * results.
 *------------------------------------------------------------------------------
 */
Browser.prototype.execute = function(
    instruction,  // Instruction to execute
    wait,         // Wait function to use, if any
    checkName,    // Name of check callback, if any
    isRecovery    // whether this execute() is part of a recovery operation
) {
    
    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 
    
    if (!isRecovery) {
        if (instruction.opcode == opcodes.OPEN) {
            this._history = [];
        }
        this._history.push({instruction: instruction, wait: wait});
    }
    
    switch(instruction.opcode) {
        case opcodes.OPEN:            
            if ((self.isBrowserRestart && self._nBrowserActions >= N_BROWSER_ACTIONS_THRESHOLD) 
                || self._nBrowserErrors > 0) {
                
                if (self._nBrowserActions >= N_BROWSER_ACTIONS_THRESHOLD) {
                    tech.debug("Restarting browser after " 
                               + self._nBrowserActions + " browser actions: " 
                               + "threshold (" 
                               + N_BROWSER_ACTIONS_THRESHOLD + ") exceeded");
                }
                if (self._nBrowserErrors > 0) {
                    tech.debug("Restarting browser after " 
                               + self._nBrowserErrors 
                               + " errors threatening its stability");
                }
                return self.restart().then(function() {
                    return executeOpen(instruction.arg1);
                })
            }
            return executeOpen(instruction.arg1);
            break;

        case opcodes.CLICK:
            return executeClick(instruction.arg1, instruction.arg2);
            break;
            
        case opcodes.BACK:
            return executeBack();
            break;
            
        case opcodes.HOVER:
            return executeHover(instruction.arg1, instruction.arg2);
            break;
        
        case opcodes.NONE:
            return executeNone();
            break;
            
        case opcodes.SCROLL:
            return executeScroll();
            
        default:
            tech.message("FATAL", "Unknown opcode " + instruction.opcode + 
                         ". Quitting");
    }
    
    function executeOpen(newUrl) {
        tech.message("DEBUG", "Opening URL " + newUrl);
        return self._driver.get(newUrl).then(finalize, handleErrors);
    }
       
    function executeClick(selector, index) {
        tech.message("DEBUG", "Finding element " + index + 
                     " to click for rule " + JSON.stringify(selector));

        return self.findElementsAtomic(selector).then(function(elements) {
            if (elements.length > index) {
                tech.message("DEBUG", "Clicking on element found");
                return elements[index].click().then(finalize, function(err) {
                    if (err instanceof webdriver.error.WebDriverError 
                        && err.code == wdErrCode.STALE_ELEMENT_REFERENCE) {
                        // Page has changed between the findElements() and 
                        // click() calls. Execute the click again.
                        tech.error("Page has changed during click. Retrying");
                        return executeClick(selector, index);
                    } else {
                        return handleErrors(err);
                    }
                });
            } else {
                // Ignore error - probably the site has changed between 
                // the generation of todoItems for this page and executing them,
                // but there's nothing the Framework can do about it, 
                // except log it and continue.
                tech.message("ERROR", "Ignoring CLICK: Less than " + (index + 1)  
                             + " elements found with selector " 
                             + JSON.stringify(selector));
            }
        });
    }

    function executeHover(selector, index) {
        tech.message("DEBUG", "Hovering over item " + index + 
                     " of rule " + JSON.stringify(selector));
        
        return self.findElementsAtomic(selector).then(function(elements) {
            if (elements.length > index) {
                return new webdriver.ActionSequence(self._driver)
                                    .mouseMove(elements[instruction.arg2])
                                    .perform()
                                    .then(finalize, handleErrors);
            } else {
                // Ignore error - probably the site has changed between 
                // the generation of todoItems for this page and executing them,
                // but there's nothing the Framework can do about it, 
                // except log it and continue.
                tech.message("ERROR", "Ignoring HOVER: Less than " + (index + 1) 
                             + " elements found with selector " 
                             + JSON.stringify(selector));
            }                    
        });
    }
    
    function executeBack() {
        tech.message("DEBUG", "Going Back");

        return self._driver.navigate().back().then(finalize, handleErrors);
    }
    
    function executeNone() {
        tech.debug("Executing NONE");
        return webdriver.promise.fulfilled().then(finalize, handleErrors);
    }
    
    function executeScroll() {
        tech.debug("Scrolling down");
        var scrollScript = "window.scrollTo(0, document.documentElement.scrollHeight - document.documentElement.clientHeight)";
        return self._driver.executeScript(scrollScript).then(function() {
            self._driver.sleep(250).then(finalize, handleErrors);
        });
    }
        
    function finalize() {
        self._nBrowserActions += 1;
        
        return self._driver.getCurrentUrl().then(function(currentUrl) {
            self._url = currentUrl;
            tech.message("INFO", "Page url: " + self._url);
            user.message("Page url: " + self._url);
            if (self._url == "about:blank") {
                self._nEmpty += 1;
                if (self._nEmpty <= N_CONNECTION_INTERRUPTED_TRIES) {
                    throw new Error("Internet connection failed");
                } else {
                    tech.fatal("Could not load page after " + N_CONNECTION_INTERRUPTED_TRIES 
                               + " attempts. Giving up");
                }
            } else {
                self._nEmpty = 0;
            }
            return webdriver.promise.all([
                executeWait(wait),
                executeCheck(checkName)
            ]);
/*
            
            self._url = currentUrl;
            tech.message("INFO", "Page url: " + self._url);
            user.message("Page url: " + self._url);
            if (wait.funcName) {
                tech.message("DEBUG", "Calling wait function " + 
                             JSON.stringify(wait));
                var waitCallback = self._waitCallbacks[wait.funcName];
                return waitCallback.call(self, wait.args).then(function() {
                    if (checkName) {
                        tech.debug("Calling check function " + checkName);
                        var checkCallback = self._checkCallbacks[checkName];
                        if (checkCallback) {
                            if (checkCallback.call(self)) {
                                tech.debug("Check succeeded");
                            } else {
                                throw new Error("Check failed");
                            }
                        } else {
                            tech.fatal("Check function " + checkName + "does not exist");
                        }
                    }
                },
                
                function(err) {
                    if (err instanceof Error && err.message.indexOf("Wait timed out") != -1) {
                        // Would be better to do some sort of recovery, but at the moment
                        // I (EHVL) don't have a good idea about what to do, as the browser
                        // is probably in an indeterminate state. There's a reason the user
                        // wants the Framework to wait...
                        // Also, one reason for the timeout is that the site is offline, and
                        // that is a fatal error anyway.
                        tech.fatal("Unable to complete browser action: Wait function timed out.");
                    }
                });
            } else {
                tech.message("DEBUG", "No wait function specified");
            }
            */
        });
    }
    
    function executeWait(wait) {
        if (wait.funcName) {
            tech.message("DEBUG", "Calling wait function " + 
                         JSON.stringify(wait));
            var waitCallback = self._waitCallbacks[wait.funcName];
            if (waitCallback) {
                return waitCallback.call(self, wait.args).thenCatch(function(err) {
                    if (err instanceof Error && err.message.indexOf("Wait timed out") != -1) {
                        // Would be better to do some sort of recovery, but at the moment
                        // I (EHVL) don't have a good idea about what to do, as the browser
                        // is probably in an indeterminate state. There's a reason the user
                        // wants the Framework to wait...
                        // Also, one reason for the timeout is that the site is offline, and
                        // that is a fatal error anyway.
                        tech.error("Unable to complete browser action: Wait function timed out.");
                        throw new Error("Wait TIMEOUT");
                    }                    
                });
            } else {
                tech.fatal("Wait function " + wait.funcName + " does not exist");
            }
        }
    }
    
    function executeCheck(checkName) {
        if (checkName) {
            tech.debug("Calling check function " + checkName);
            var checkCallback = self._checkCallbacks[checkName];
            if (checkCallback) {
                return checkCallback.call(self).then(function(isOk) {
                    if (isOk) {
                        tech.debug("Check succeeded");
                    } else {
                        throw new Error("Check failed");
                    }
                });
            } else {
                tech.fatal("Check function " + checkName + " does not exist");
            }
        }
    }
    
    function handleErrors(err) {
        if (err instanceof webdriver.error.WebDriverError 
            && err.code == wdErrCode.TIMEOUT) {
            // Note: timeout errors are most likely due to a phantomjs bug,
            // in which the "page loading" flag is erroneously left on.
            // So we log the error and continue. Since this is a phantomjs bug
            // we trigger a browser restart at the earliest opportunity.
            tech.error("Phantomjs TIMEOUT error");
            self._nBrowserErrors += 1;

            // Simulate pressing the "stop page loading" button
            // (this seems to help in some cases)
            return self._driver.executeScript("return window.stop()").then(function() {
                return finalize();
            });
        } else {
            throw err;
        }
    }
}

/*------------------------------------------------------------------------------
 * Browser.getSource()
 *
 * returns parsed page source of the page currently visible in the browser
 *------------------------------------------------------------------------------
 */
Browser.prototype.getSource = function() {
    
    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 
    
    tech.debug("Obtaining page source");
    return self._driver.getPageSource().then(function(html) {
        return cheerio.load(html);
    });
}

/*------------------------------------------------------------------------------
 * Browser.getUrl()
 *
 * Returns URL of the current page
 *------------------------------------------------------------------------------
 */
Browser.prototype.getUrl = function() {
    
    return this._url;
}

/*------------------------------------------------------------------------------
 * Browser.findElements()
 *
 * find elements on the page. Wrapper around selenium findElements()
 * Returns a promise which will be resolved with the elements found.
 *
 * Note: throws STALE_ELEMENT_REFERENCE when a page changes between the
 * driver.findElements() and isDisplayed() calls.
 *
 * edit EHVL 20160217: added condition. Depending on condition, all
 * elements are returned, or only the (in)visible ones. If condition is omitted
 * or contains any value other than "all" or "invisible" only the visible
 * ones are returned.
 *------------------------------------------------------------------------------
 */
Browser.prototype.findElements = function(selector, condition) {
    
    tech.debug("Finding elements with selector: " + JSON.stringify(selector));
    if (typeof(condition) === "undefined") {
        condition = "visible";
    }
    tech.debug("Visibility condition: " + condition);
    
    var allElements = this._driver.findElements(selector);
    if (condition === "all") {
        return allElements;
    }
    return webdriver.promise.filter(allElements, function(element) {
        return element.isDisplayed().then(function(isVisible) {
            if (condition === "invisible") {
                return !isVisible;
            } else {
                return isVisible;
            }
        });
    });
}    
    
/*------------------------------------------------------------------------------
 * Browser.findElementsAtomic()
 *
 * Atomic version of Browser.findElements(). Catches site changes for which
 * Browser.findElements() throws STALE_ELEMENT_REFERENCE by repeating the search
 *------------------------------------------------------------------------------
 */
Browser.prototype.findElementsAtomic = function(selector, condition) {
    
    var self = this;
    
    return this.findElements(selector, condition).thenCatch(function(err) {
        if (err instanceof webdriver.error.WebDriverError
            && err.code == wdErrCode.STALE_ELEMENT_REFERENCE) {
            tech.debug("Website changed during findElements. Retrying");
            return self.findElementsAtomic(selector);
        } else {
            throw err;
        }
    });
}        

/*------------------------------------------------------------------------------
 * Browser.getReloadDelay()
 *
 * returns the delay (in milliseconds) for page reload attempts
 *------------------------------------------------------------------------------
 */
Browser.prototype.getReloadDelay = function() {
    return RELOAD_DELAY_MS;
}

Browser.prototype.isFullyScrolled = function() {
    
    tech.debug("Checking if fully scrolled");
    var isFullyScrolledScript = 
        "return { scrollHeight: document.documentElement.scrollHeight, scrollY: window.scrollY, clientHeight: document.documentElement.clientHeight}";
        //"return { scrollHeight: document.documentElement.scrollHeight, scrollTop: document.documentElement.scrollTop, clientHeight: document.documentElement.clientHeight}";
        //"return document.documentElement.scrollHeight - document.documentElement.scrollTop === document.documentElement.clientHeight";
    
    return this._driver.executeScript(isFullyScrolledScript).then(function(scrollParms) {
        //tech.debug("Fully scrolled: " + isScrolled);
        tech.debug("scrollHeight: " + scrollParms.scrollHeight);
        tech.debug("scrollY: " + scrollParms.scrollY);
        tech.debug("clientHeight: " + scrollParms.clientHeight);
        tech.debug(scrollParms.scrollHeight - scrollParms.scrollY);
        console.log(scrollParms);
        console.log(typeof(scrollParms.scrollHeight));
        console.log(typeof(scrollParms.scrollY));
        
        var result = scrollParms.scrollHeight - scrollParms.scrollY == scrollParms.clientHeight;
        tech.debug("Fully scrolled: " + result);
        return result;
    })
}

Browser.prototype.getScrollPosition = function() {
    
    tech.debug("Getting scroll position");
    var getScrollPosScript = 
        "return { scrollY: window.scrollY, maxScrollY: document.documentElement.scrollHeight - document.documentElement.clientHeight }";

    return this._driver.executeScript(getScrollPosScript);
}
    
module.exports = Browser;
