/*=============================================================================
 * 
 * Robot Framework 4
 * Extractor object
 *
 * (c) 2015 Statistics Netherlands
 *
 * Revision history:
 * 02 march    2015: EHVL: split off from framework.js
 * 02 october  2015: EHVL: fixed bug in extractAttribute() if attribute 
 *                         is absent
 *
 *=============================================================================
 */

var moment = require("moment");
var fs = require("fs");
var csv = require("fast-csv");
var path = require("path");
var _ = require("underscore");

var tech = require("./levellogger");
var user = require("./logger");
var EXCEL_CSV_NL = require("./outputformat");
var columnCounts = require("./columncounts");

/*------------------------------------------------------------------------------
 * Extractor constructor
 *------------------------------------------------------------------------------
 */
function Extractor(
    extractionRules,  // The extraction rules as specified by the user
    robotName,        // Name of the robot
    robotVersion,     // Version of the robot
    dataDir           // Output directory
) {
    
    var i;

    // Initialize object properties
    this._extractRules = {};
    this._robotName = robotName;
    this._robotVersion = robotVersion;
    this._obsDate = moment().format("YYYYMMDD");
    this._callbacks = {
        extractXmlValue: function($, item) {
            return $(item).text().trim().replace(/\s+/g, " ");
        },
        extractAttribute: function($, item, attrName) {
            var attrValue = $(item).attr(attrName);
            if (typeof(attrValue) === "undefined") attrValue = "";
            return attrValue.trim();
        }
    }

    // Parse extraction rules
    tech.message("DEBUG", "Parsing extraction rules");

    var ruleNames = Object.keys(extractionRules);        
    for (var i = 0; i < ruleNames.length; i++) {
        var ruleName = ruleNames[i];
        var inputRule = extractionRules[ruleName];

        if (!inputRule.hasOwnProperty("cssSelector")) {
            tech.message("FATAL", "Required property 'cssSelector' not " +
                         "present in extraction rule " + ruleName);
        }

        this._extractRules[ruleName] = {
            itemSelector: inputRule.cssSelector,
            properties: []
        };  
        var parsedRule = this._extractRules[ruleName]
        var varNames = Object.keys(inputRule);

        for (var j = 0; j < varNames.length; j++) {
            var varName = varNames[j];
            
            if (varName != "cssSelector") {
                // Skip cssSelector here, as it is not an output variable name.
                var varParms = inputRule[varName];
                
                parsedRule.properties.push(
                    {
                        varName: varName,
                        selector: varParms.cssSelector || "",
                        attrName: varParms.attribute || "",
                        operation: varParms.operation
                    }
                );
            }
        }
    }

    // initialize output files, one file for each extraction context
    this._outputFiles = {};
    for (var extractionContext in this._extractRules) {
        var newFile = csv.format(EXCEL_CSV_NL);
        var fileName = this._obsDate
            + this._robotName
            + extractionContext
            + this._robotVersion + ".csv";

        newFile.pipe(fs.createWriteStream(path.join(dataDir, fileName)));
        this._outputFiles[extractionContext] = newFile;
        tech.message("DEBUG", "Output file " + fileName + " initialized");
    }
        
    // Initialize column counts
    this._columnCounts = {};
    for (var extractionContext in this._extractRules) {
        this._columnCounts[extractionContext] = new columnCounts();
    }
}

/*------------------------------------------------------------------------------
 * Extractor.registerCallbacks()
 *
 * Add custom data extraction callbacks to the extractor object
 *------------------------------------------------------------------------------
 */
Extractor.prototype.registerCallbacks = function(
    newCallbacks  // The custom extraction callbacks to add
) {

    _.extend(this._callbacks, newCallbacks)
}

/*------------------------------------------------------------------------------
 * Extractor.processPage()
 *
 * Extract data from a parsed HTML page
 *------------------------------------------------------------------------------
 */
Extractor.prototype.processPage = function(
    $,        // Parsed HTML page
    todoItem  // Current todo item
) {

    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 
    var extractionContext = todoItem.extractionContext;
    
    if (!extractionContext || 
        !self._extractRules.hasOwnProperty(extractionContext)) {
        
        if (extractionContext) {
            // Ignore - but log - invalid extraction context
            tech.message("INFO", "Unknown extraction context " + 
                         extractionContext + " ignored");
        }
        return;
    }
    
    // Find items on the page
    var extractRule = self._extractRules[extractionContext];
    var items = $(extractRule.itemSelector);
    
    tech.message("INFO", "Number of items found on page: " + items.length);
    user.message("Number of items found on page: " + items.length);
    
    var properties = extractRule.properties;
    var outputfile = self._outputFiles[extractionContext];
    var columnCounts = self._columnCounts[extractionContext];
    
    // Loop over all items found
    items.each(function(j, item) {
        
        var i;
        var record = {};
        var useItem;
        var callback;
        
        // Loop over all item properties 
        for (i=0; i<properties.length; i++) {
            var property = properties[i];
                
            if (property.selector.length == 0) {
                /* If property is empty, use the item as a whole as property.
                 * Used mainly to select an attribute of the tag containing
                 * the item */
                useItem = item;
            } else {
                useItem = $(property.selector, item);
            }
            if (self._callbacks.hasOwnProperty(property.operation)) {
                callback = self._callbacks[property.operation];
                value = callback($, useItem, property.attrName);
                record[property.varName] = value;
            } else {
                record[property.varName] = "NA";
            }
        }
        
        // Add navigation variables and metadata to the output record
        _.extend(record, todoItem.vars);
        record.Date = self._obsDate;
        record.RobotName = self._robotName;
        record.RobotVersion = self._robotVersion;
        
        outputfile.write(record);
        columnCounts.add(record);
    });
}

/*------------------------------------------------------------------------------
 * Extractor.generateReports()
 *
 * Generates empty column count (the number of completely empty columns in the 
 * output generated) and single value column count (the number of columns in the
 * output with exactly one non-empty value).
 *------------------------------------------------------------------------------
 */
Extractor.prototype.generateCounts = function() {
    
    var empty  = 0;
    var single = 0;
    
    for (var extractionContext in this._extractRules) {
        empty  += this._columnCounts[extractionContext].countEmpty();
        single += this._columnCounts[extractionContext].countSingle();
    }
    
    return {
        emptyColumnCount:       empty,
        singleValueColumnCount: single
    }
}
        
module.exports = Extractor;
