/*=============================================================================
 * 
 * Robot Framework 4
 * Levellogger object
 *
 * (c) 2015 Statistics Netherlands
 *
 * Revision history:
 * 22 january 2016: EHVL: creation of this comment block - 
 *                        previous code changes have not been tracked
 *
 *=============================================================================
 */

var Logger = require("./logger-prototype");

const LOG_LEVEL_NAMES = [ "DEBUG", "INFO", "WARNING", "ERROR", "FATAL" ];
const DEFAULT_LOG_LEVEL_NAME = "INFO";

var leftAlignLevels = leftAlignArray(LOG_LEVEL_NAMES);

/*------------------------------------------------------------------------------
 * leftAlignArray()
 *
 * Left-aligns the contents of a text array: the resulting array has the same
 * contents of the input array, but each string in the result has the same length
 * as the longest string in the input, padded on the right with spaces.
 *------------------------------------------------------------------------------
 */
function leftAlignArray(input) {
    
    var maxLength = -1;
    for (var i = 0; i < input.length; i++) {
        if (input[i].length > maxLength) {
            maxLength = input[i].length;
        }
    }
    
    // Obtain a string of maxLength space characters
    var padding = Array(maxLength + 1).join(" ");
    
    var output = Array(input.length);
    for (i = 0; i < input.length; i++) {
        output[i] = (input[i] + padding).substring(0, maxLength);
    }
    
    return output;
}

/*------------------------------------------------------------------------------
 * getLevel()
 *
 * Obtains a numeric value for the log level, based on the position in the 
 * LOG_LEVEL_NAMES array.
 *------------------------------------------------------------------------------
 */
function getLevel(levelName) {
    var level = LOG_LEVEL_NAMES.indexOf(levelName);
    if (level === -1) {
        throw new Error("Invalid log level " + levelName);
    }
    return level;
}
    
/*------------------------------------------------------------------------------
 * Levellogger constructor
 * 
 * Levellogger inherits Logger, with the log level as extra property.
 *------------------------------------------------------------------------------
 */
function Levellogger() {
    this._level = LOG_LEVEL_NAMES.length;
    Logger.call(this);
}
    
/* Perform inheritance of the Logger prototype */
Levellogger.prototype = Object.create(Logger.prototype);
Levellogger.prototype.constructor = Levellogger;

/*------------------------------------------------------------------------------
 * Levellogger.open()
 * 
 * Opens a Levellogger log file and sets the log level
 *------------------------------------------------------------------------------
 */
Levellogger.prototype.open = function(filename, levelname) {
    if (typeof levelname !== "string") {
        levelname = DEFAULT_LOG_LEVEL_NAME;
    }
    this._level = getLevel(levelname);
    Logger.prototype.open.call(this, filename);
}

/*------------------------------------------------------------------------------
 * Levellogger.close()
 * 
 * Closes a Levellogger log file
 *------------------------------------------------------------------------------
 */
Levellogger.prototype.close = function() {
    this._level = LOG_LEVEL_NAMES.length;
    Logger.prototype.close.call(this);
}

/*------------------------------------------------------------------------------
 * Levellogger.message()
 * 
 * Logs a message to a Levellogger log file, based on the log level 
 * of the message and the log file. Exits the program on a FATAL error.
 *------------------------------------------------------------------------------
 */
Levellogger.prototype.message = function(levelname, text) {
    level = getLevel(levelname);
    if (level >= this._level) {
        Logger.prototype.message.call(
            this, leftAlignLevels[level] + " " + text);
    }
    if (level == LOG_LEVEL_NAMES.length - 1) { // FATAL
        console.error(text);
        console.trace();
        process.exit();
    }
}

/*------------------------------------------------------------------------------
 * Levellogger.getLogLevelName()
 * 
 * Returns the log level name of the Levellogger log file.
 *------------------------------------------------------------------------------
 */
Levellogger.prototype.getLogLevelName = function() {
    return LOG_LEVEL_NAMES[level];
}

// Generate convenience functions: Levellogger.info(...) as shorthand for 
// Levellogger.message("INFO", ...) and so on.
for (var i = 0; i < LOG_LEVEL_NAMES.length; i++) {
    var func = LOG_LEVEL_NAMES[i].toLowerCase();
    eval(
        "Levellogger.prototype." + func + " = function(text) {" +
        "    this.message('" + LOG_LEVEL_NAMES[i] + "', text);" + 
        "}"
    );
}

module.exports = Levellogger;
