/*=============================================================================
 * 
 * Robot Framework 4
 * Navigator object
 *
 * (c) 2015 Statistics Netherlands
 *
 * Revision history:
 * 02 march    2015: EHVL: split off from framework.js
 *
 *=============================================================================
 */

var _ = require("underscore");
var async = require("async");
var url = require("url");
var cheerio = require("cheerio");
var webdriver = require("selenium-webdriver");

var user = require("./logger");
var tech = require("./levellogger");
var opcodes = require("./opcodes");
var wait = require("./wait");
var TodoItem = require("./todoitem");
var scroll = require("./scroll");

var wdErrCode = webdriver.error.ErrorCode;

/*------------------------------------------------------------------------------
 * Navigator constructor
 *
 * This constructor initializes a Navigator object by parsing the 
 * navigationRules and backRules configuration tables
 *------------------------------------------------------------------------------
 */
function Navigator(
    userNavRules,     // user-specified navigation rules
    userBackRules,    // user-specified back rules
    browser           // a browser object
) {

    // Initialize object properties
    this._navRules = {};
    this._backRules = {};
    this._vars = null;
    this._backRulename = "";
    this._browser = browser;
    this._urlCallbacks = {
        getHref: function(element, currentUrl) {
            return element.getAttribute("href").then(function(newUrl) {
                return url.resolve(currentUrl, newUrl);
            });
        }
    };
    this._textCallbacks = {
        getElementText: function(element) {
            return element.getText().then(function(text) {
                return text.trim();
            });
        },
        getElementTextAll: function(element) {
            return element.getInnerHtml().then(function(html) {
                var $ = cheerio.load(html);
                            
                return $.root().text().trim();
            }); 
        },
        getElementTextOnly: function(element) {
            return element.getInnerHtml().then(function(html) {
                var $ = cheerio.load(html);
                            
                return $.root().contents().filter(function() {
                    return this.type === "text";
                }).text().trim();
            }); 
        } 
    };

    // Parse forward navigation rules
    tech.message("DEBUG", "Parsing navigation rules");
    var navContexts = Object.keys(userNavRules);
    for (var i = 0; i < navContexts.length; i++) {
        var navContext = navContexts[i];
        var userNavRule = userNavRules[navContext];
        
        // Deal with  navigation contexts with a single rule, in which case
        // no array is required.
        if (!Array.isArray(userNavRule)) {
            userNavRule = [ userNavRule ];
        }
        
        var navRule = [];
        this._navRules[navContext] = navRule;
        for (var j = 0; j < userNavRule.length; j++) {
            navRule[j] = {
                opcode: opcodes.parse(userNavRule[j].action, 
                                      "in navigation context " + navContext + 
                                      ", rule " + (j+1)),
                selector: userNavRule[j].xpathSelector ? 
                    { xpath: userNavRule[j].xpathSelector} : 
                    { css: userNavRule[j].cssSelector },
                selectorCondition: userNavRule[j].selectorCondition,
                urlOperation: userNavRule[j].urlOperation || null,
                textOperation: userNavRule[j].textOperation || null,
                varName: userNavRule[j].navigationVariable || null,
                newNavigationContext: userNavRule[j].navigationContext || null,
                newExtractionContext: userNavRule[j].extractionContext || null,
                backRuleName: userNavRule[j].backRule || null,
                wait: wait.parse(userNavRule[j].wait, " in navigation rule " + 
                                j + " of navigation context " + navContext),
                check: userNavRule[j].checkOperation || null
            };
            if (navRule[j].opcode == opcodes.SCROLL || 
                (navRule[j].opcode == opcodes.NONE && !userNavRule[j].xpathSelector && !userNavRule[j].cssSelector)) {
                navRule[j].selector = { scroll: scroll.parse(userNavRule[j].scrollSelector) };
            }
        }
    }
     
    // Parse backward navigation rules
    tech.message("DEBUG", "Parsing back rules");
    var backRuleNames = Object.keys(userBackRules);
    for (var i = 0; i < backRuleNames.length; i++) {
        var backRuleName = backRuleNames[i];
        var userBackRule = userBackRules[backRuleName];
        
        if (!Array.isArray(userBackRule)) {
            userBackRule = [ userBackRule ];
        }
        
        var backRule = [];
        this._backRules[backRuleName] = backRule;
        for (var j = 0; j < userBackRule.length; j++) {
            backRule[j] = {
                opcode: opcodes.parse(userBackRule[j].action, 
                                      "in back rule block " + backRuleName + 
                                      ", rule " + (j+1)),
                wait: wait.parse(userBackRule[j].wait, 
                                 " in back rule block " + backRuleName + 
                                 ", rule " + (j+1)),
                check: userBackRule[j].checkOperation
            };
            switch(backRule[j].opcode) {
                case opcodes.OPEN:
                    backRule[j].arg = userBackRule[j].url;
                    break;
                case opcodes.CLICK:
                case opcodes.HOVER:
                    backRule[j].arg = userBackRule[j].xpathSelector ?
                        { xpath: userBackRule[j].xpathSelector } :
                        { css: userBackRule[j].cssSelector };
                    break;
                case opcodes.BACK:
                case opcodes.DUPLICATE:
                    backRule[j].arg = null;
                    break;
                case opcodes.REPEAT: 
                    backRule[j].arg = userBackRule[j].repeatAmount;
                    break;
            }
        }
    }
}

/*------------------------------------------------------------------------------
 * Navigator.registerUrlCallbacks()
 *
 * Add custom navigation callbacks to the Navigator object
 *------------------------------------------------------------------------------
 */
Navigator.prototype.registerUrlCallbacks = function(newUrlCallbacks) {

    _.extend(this._urlCallbacks, newUrlCallbacks)
}

/*------------------------------------------------------------------------------
 * Navigator.registerTextCallbacks()
 *
 * Add custom navigation callbacks to the Navigator object
 *------------------------------------------------------------------------------
 */
Navigator.prototype.registerTextCallbacks = function(newTextCallbacks) {

    _.extend(this._textCallbacks, newTextCallbacks)
}

/*------------------------------------------------------------------------------
 * Navigator.getForwardItems()
 *
 * Obtains a list of all new todoItems generated from the navigation rules
 * applicable to the current page. The function returns a promise which is
 * resolved with an array containing all these new todoItems.
 *
 * Each new todoItem is paired up with the backward todoItems (generated by
 * Navigator.getBackwardItems) generated by the back rule associated with the
 * todoItem.
 *------------------------------------------------------------------------------
 */
Navigator.prototype.getForwardItems = function(
    currentUrl,   // Url of the current page
    todoItem      // current todo item
) {

    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 
    
    if (!todoItem.navigationContext || 
        !self._navRules.hasOwnProperty(todoItem.navigationContext)) {
        
        // Non-existing navigation contexts are ignored and logged
        if (todoItem.navigationContext) {
            tech.error("Unknown navigation context " + 
                       todoItem.navigationContext + " ignored");
        }
        // No todoItems found: return a promise which resolves with an empty array. 
        return webdriver.promise.fulfilled([]);
    }

    // Find all navigation elements belonging to all navigation rules for the 
    // current navigation context and construct new todoItems for each of them.
    var navRules = self._navRules[todoItem.navigationContext];

    return webdriver.promise.map(navRules, function(navRule) {
        return performSelection(navRule.selector, navRule.selectorCondition).then(function(elements) {
            return webdriver.promise.map(elements, function(element, i) {
                // Execute urlOperation and textOperation for the current element
                var nullFunc = function() { 
                    return webdriver.promise.fulfilled(null); 
                }
                var getUrl = self._urlCallbacks[navRule.urlOperation];
                if (!getUrl) getUrl = nullFunc;
                var getText = self._textCallbacks[navRule.textOperation];
                if (!getText) getText = nullFunc;

                return webdriver.promise.all([
                    getUrl(element, currentUrl),
                    getText(element)
                ]).then(function(data) {
                    var arg1, arg2;
                    // Construct new todoItem
                    switch(navRule.opcode) {
                        case opcodes.OPEN:
                            arg1 = data[0];
                            break;
                        case opcodes.CLICK:
                        case opcodes.HOVER:
                            arg1 = navRule.selector;
                            arg2 = i;
                            break;
                        case opcodes.NONE:
                        case opcodes.SCROLL:
                            break;
                        default:
                            tech.message("FATAL", "Invalid opcode " + 
                                         navRule.opcode + 
                                         " in navigation context " + 
                                         todoItem.navigationContext);
                    }
                    var newTodoItem = new TodoItem(navRule.newNavigationContext, 
                                                   navRule.newExtractionContext,
                                                   todoItem.vars,
                                                   navRule.wait,
                                                   navRule.check,
                                                   navRule.opcode,
                                                   arg1,
                                                   arg2);
                    newTodoItem.setVariable(navRule.varName, data[1]);
                    backTodoItems = self.getBackwardItems(navRule.backRuleName, 
                                                          todoItem.vars);
                    return { forward: newTodoItem, backward: backTodoItems };
                });
            });
        });
    }).then(function(todoItems) {
        // Make a simple array from an array of arrays
        result = Array.prototype.concat.apply([], todoItems);
        tech.message("INFO", "Number of navigation items found: " + result.length);
        user.message("Number of navigation items found: " + result.length);
        
        return result;
    }, function(err) {
        if (err instanceof webdriver.error.WebDriverError) {
            if (err.code == wdErrCode.STALE_ELEMENT_REFERENCE) {
                // Page has changed between findElements() and 
                // getUrl() / getText() calls. Redo the entire navigation 
                // to get the correct navigation items
                tech.error("Page changed during navigation. Doing navigation again");
                return Navigator.prototype.getForwardItems.call(self, currentUrl, todoItem);
            } else {
                throw err;
            }
        } else {
            throw err;
        }
    });
    
    function performSelection(selector, condition) {
        
        if (selector.scroll) {
            return self._browser.getScrollPosition().then(function(scrollParms) {
                if (scroll.compare(scrollParms.scrollY, scrollParms.maxScrollY, selector.scroll)) {
                    return [ true ];
                } else {
                    return [];
                }
            });
        } else {
            return self._browser.findElements(selector, condition);
        }
    }
}

/*------------------------------------------------------------------------------
 * Navigator.getBackwardItems()
 *
 * Obtain a list of all todo items generated by the indicated back rule.
 * Note that this is a simple, synchronous function.
 *------------------------------------------------------------------------------
 */
Navigator.prototype.getBackwardItems = function(
    backRuleName,         // Name of the back rule chain to use
    vars                  // Navigation variables
) {

    var result = [];

    // Ignore non-existing backrules
    if (backRuleName && this._backRules.hasOwnProperty(backRuleName)) {

        // Simple synchronous loop over all back rules with the current back rule name
        var backRuleChain = this._backRules[backRuleName];
        for (var i =  0; i < backRuleChain.length; i++) {
            var backrule = backRuleChain[i];
            var newTodoItem = new TodoItem(null, 
                                           null, 
                                           vars, 
                                           backrule.wait, 
                                           backrule.check,
                                           backrule.opcode, 
                                           backrule.arg);

            result.push(newTodoItem);
        }
    }
    return result;
}

/*------------------------------------------------------------------------------
 * Navigator.getNavigationVars()
 *
 * Return an object containing all navigation variables. Each navigation 
 * variable gets the value "NA", which means "empty" in R.
 *------------------------------------------------------------------------------
 */
Navigator.prototype.getNavigationVars = function() {
    
    var result = {};
    var i;

    for (var navRuleContextName in this._navRules) {
        var navRuleContext = this._navRules[navRuleContextName];
        
        for (i=0; i<navRuleContext.length; i++) {
            var varName = navRuleContext[i].varName;
            if (varName) {
                
                // Comply with the standard for empty fields in R
                result[varName] = "NA";
            }
        }
    }
    
    return result;
}

module.exports = Navigator;
