/*=============================================================================
 * 
 * Robot Framework 4
 * Robot object
 *
 * (c) 2015 Statistics Netherlands
 *
 * Revision history:
 * 02 march    2015: EHVL: split off from framework.js
 * 14 april    2015: EHVL: removed () from a function name in a setImmediate()
 *                         this erroneously executed the function and registered
 *                         the /result/ as a callback, in stead of the function 
 *                         itself.
 * 01 september 2015: EHVL: new startUrls syntax; parsing changed
 *
 *=============================================================================
 */

var async = require("async");
var csv = require("fast-csv");
var moment = require("moment");
var path = require("path");
var fs = require("fs");
var webdriver = require("selenium-webdriver");

var user = require("./logger");
var tech = require("./levellogger");
var Extractor = require("./extractor");
var Navigator = require("./navigator");
var Browser = require("./browser");
var TodoItem = require("./todoitem");
var LinkedList = require("./linkedlist");
var wait = require("./wait");
var opcodes = require("./opcodes");
var EXCEL_CSV_NL = require("./outputformat");

var wdErrCode = webdriver.error.ErrorCode;

/*------------------------------------------------------------------------------
 * Robot constructor
 *------------------------------------------------------------------------------
 */
function Robot(
    specific,       // The site-specific code
    dataDir,        // Directory in which the output is written
    logDir,         // Directory in which the category count is written
    browserLogName, // Filename of the browser log 
    browserLogLevel // Browser log level
) {

    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 

    // Asynchronically initialize browser. 
    // Done first because it takes the most time.
    self._browser = new Browser(specific.variables.userAgent, 
                                path.join(logDir, browserLogName),
                                browserLogLevel, 
                                specific.variables.isBrowserRestart);
    self._browser.registerWaitCallbacks(specific.waitCallbacks);
    self._browser.registerCheckCallbacks(specific.checkCallbacks);
        
    // Initialize navigation
    self._navigator = new Navigator(specific.navigationRules, 
                                    specific.backRules, 
                                    self._browser);
    self._navigator.registerUrlCallbacks(specific.urlCallbacks);
    self._navigator.registerTextCallbacks(specific.textCallbacks);

    // Initialize data extraction
    self._extractor = new Extractor(specific.extractionRules, 
                                    specific.variables.robotName, 
                                    specific.variables.version, 
                                    dataDir);
    self._extractor.registerCallbacks(specific.extractionCallbacks);

    // Initialize Todo list
    self._todoList = new LinkedList();

    var emptyNavigationVars = self._navigator.getNavigationVars();
    tech.message("INFO", "Navigation variables: " + 
                 Object.keys(emptyNavigationVars));

    // Add start urls to todo list
    var startUrls = specific.startUrls;
    
    // Determine if startVariable is present
    var startVar = startUrls.startVariable;
    if (!startUrls.hasOwnProperty("startVariable")) {
        startVar = false;
    }

    // Loop over all start URLs
    var startValues = Object.keys(startUrls);
    for (var i = 0; i < startValues.length; i++) {
        
        var startValue = startValues[i];
        if (startValue === "startVariable") {
            continue;
        }
        
        var startUrl = startUrls[startValue];
        tech.message("DEBUG", "Adding start URL " + startUrl.url + 
                     " to the todo list");
        if (startVar) {
            emptyNavigationVars[startVar] = startValue;
        }
        var waitFunc = wait.parse(startUrl.wait, " in start URL " + startValue);
        var todoItem = new TodoItem(startUrl.navigationContext || null,
                                    startUrl.extractionContext || null,
                                    emptyNavigationVars,
                                    waitFunc,
                                    "",
                                    opcodes.OPEN,
                                    startUrl.url);
        self._todoList.insertAfter(todoItem);
    }
        
    // Initialize counts file
    var countsFilename = moment().format("YYYYMMDD")
        + specific.variables.robotName
        + "counts"
        + ".csv";

    self._countsFilename = path.join(logDir, countsFilename);
}

/*------------------------------------------------------------------------------
 * Robot._navigate()
 *
 * Updates the todo list with new todo items found by the navigator. This code
 * contains the control logic necessary to keep the state of the browser 
 * consistent with the state of the robot.
 *------------------------------------------------------------------------------
 */
Robot.prototype._navigate = function(
    currentUrl,   // Url of the current page
    todoItem,     // Current todo item
    curPos        // Position of current todo item in todo list
) {
     
    // standard workaround, avoids problems involving "this" in inner functions
    var self = this;
    
    /* Obtain positions to insert new todo items obtained by the navigator.
     * Some are inserted directly after the current item, while others are 
     * inserted at the end. The new items are inserted back to front, which has
     * the advantage that the insert positions do not need to be updated after
     * each insertion. */
    var lastPos = self._todoList.getLast();
    // Use the navigator to obtain all new todo items
    return this._navigator.getForwardItems(currentUrl, todoItem).then(function(newTodoItems) {
        for (i = newTodoItems.length - 1; i >= 0; i--) {
            newTodoItem = newTodoItems[i].forward;
            if (newTodoItem.instruction.opcode == opcodes.OPEN) {
                /* Todo items that OPEN a new url are appended to the end of the
                 * todo list because they disturb the current state of the browser,
                 * and don't need it anyway. */
                self._todoList.insertAfter(newTodoItem, lastPos);

            } else {
                /* Other todo items (those doing something with the current page)
                 * are inserted right after the current todo item because they 
                 * depend on the current state of the browser */

                /* Check if the next item on the todo list depends on the current
                 * browser state. Since we are inserting the new todo item right
                 * in front of this item, and the new todo item might disturb the
                 * current browser state, we need to "repair" the browser state if
                 * it does. This is done by means of one or more "back" actions */
                var nextPos = self._todoList.getNext(curPos);

                if (nextPos != null) {
                    var peekItem = self._todoList.getData(nextPos);

                    if (peekItem.instruction.opcode != opcodes.OPEN) {
                        var backTodoItems = newTodoItems[i].backward;
                        for (j = backTodoItems.length - 1; j >= 0; j--) {
                            var backTodoItem = backTodoItems[j];

                            // Deal with pseudo-opcodes
                            if (backTodoItem.instruction.opcode == opcodes.DUPLICATE) {

                                // Add a copy of the new todo item to the todo list
                                backTodoItem = new TodoItem(null, 
                                                            null,
                                                            backTodoItem.vars,
                                                            newTodoItem.wait,
                                                            newTodoItem.check,
                                                            newTodoItem.instruction.opcode,
                                                            newTodoItem.instruction.arg1,
                                                            newTodoItem.instruction.arg2);
                                self._todoList.insertAfter(backTodoItem, curPos);

                            } else if (backTodoItem.instruction.opcode == opcodes.REPEAT) {

                                /* Add copies of the most recently processed todo items
                                 * to the todo list, so that the same actions will be 
                                 * performed again. This is one (quite expensive but sometimes
                                 * necessary) way to get the same browser state as before. */
                                var repeatPos = curPos;
                                for (var k = backTodoItem.instruction.arg1; k > 0; k--) {                                
                                    var oldItem = self._todoList.getData(repeatPos);
                                    backTodoItem = new TodoItem(null,
                                                                null,
                                                                oldItem.vars,
                                                                oldItem.wait,
                                                                oldItem.check,
                                                                oldItem.instruction.opcode,
                                                                oldItem.instruction.arg1,
                                                                oldItem.instruction.arg2);
                                    self._todoList.insertAfter(backTodoItem, curPos);
                                    repeatPos = self._todoList.getPrev(repeatPos);
                                }
                            } else {
                                self._todoList.insertAfter(backTodoItem, curPos);
                            }
                        }
                    }
                }
                /* Add the new todo item in front of the 
                 * newly added back items (if any) */
                self._todoList.insertAfter(newTodoItem, curPos);
                if (nextPos == null) { 
                    // The new item added is the (new) last item on the list
                    lastPos = self._todoList.getLast();
                }
            }
        }
    });
}

/*------------------------------------------------------------------------------
 * Robot._processItem()
 * 
 * Processes one todo item.
 *------------------------------------------------------------------------------
 */
Robot.prototype._processItem = function(
    currentPos,  // position in the todo list of todo item to be processed
    callback     // Function to be called once processing is complete
) {
    
    // standard workaround, avoids problems involving "this" in inner functions
    var self = this; 
    var todoItem = self._todoList.getData(currentPos);

    // Execute the browser instruction of the todo item
    self._browser.execute(todoItem.instruction, todoItem.wait, todoItem.check).then(function() {
        tech.message("DEBUG", "Extraction context: " +
                     todoItem.extractionContext);
        tech.message("DEBUG", "Navigation context: " + 
                     todoItem.navigationContext);
        tech.message("INFO", "Navigation variable content: " + 
                     JSON.stringify(todoItem.vars));

        // Perform navigation 
        return self._navigate(self._browser.getUrl(), todoItem, currentPos);
    
    }).then(function() {        

        // Extract data from the resulting page, if necessary
        if (todoItem.extractionContext) {
            return self._browser.getSource().then(function($) {
                self._extractor.processPage($, todoItem);
            });
        }
    }).then(callback, function(err) {
        if (err instanceof webdriver.error.WebDriverError) {
            if (err.code == wdErrCode.UNKNOWN_ERROR) {
                tech.error("Browser encountered 'unknown error'");
                self._browser.recover().then(function() {
                    Robot.prototype._processItem.call(self, currentPos, callback);
                });
            } else {
                console.error(err.stack);
                tech.fatal("Uncaught selenium error: " + err);
            }
        } else if (err instanceof Error) {
            if (err.message.indexOf("ECONNREFUSED") != -1
                || err.message.indexOf("EADDRINUSE") != -1
                || err.message.indexOf("Server terminated early") != -1
                || err.message == "Check failed"
                || err.message == "Internet connection failed"
                || err.message == "Wait TIMEOUT") {
                tech.error("Recovering from system error: " + err.message);
                /*
                if (err.message.indexOf("ECONNREFUSED") != -1) {
                    tech.error("Browser crashed");
                } else if (err.message == "Check failed") {
                    tech.error("Check failed");
                } else if (err.message == "Internet connection failed") {
                    tech.error("Connection failure");
                }
                */
                self._browser.recover().then(function() {
                    var delay = 0;
                    if (err.message == "Internet connection failed") {
                        delay = self._browser.getReloadDelay();
                    }
                    setTimeout(function() {
                        Robot.prototype._processItem.call(self, currentPos, callback);
                    }, delay);
                });
            } else {
                console.error(err.stack);
                tech.fatal("Uncaught system error: " + err);
            }
        } else {
            console.error(err.stack);
            tech.fatal("Uncaught error: " + err);
        }
    });
}

/*------------------------------------------------------------------------------
 * Robot.processItems()
 * 
 * Processes all items on the todo list.
 *------------------------------------------------------------------------------
 */
Robot.prototype.processItems = function() {

    var self = this;

    this._browser.start().then(function() {
        var currentPos = self._todoList.getFirst();

        async.whilst(
            function() { return currentPos != null },
            function(next) {
                self._processItem(currentPos, function() {

                    currentPos = self._todoList.getNext(currentPos);
                    setImmediate(next);
                });
            },
            function() {
                tech.message("INFO", "Run complete");
                user.message("Run complete");

                var catCount = self._getCategoryCount();
                var columnCounts = self._extractor.generateCounts()

                self._saveCounts(
                    catCount,
                    columnCounts.emptyColumnCount,
                    columnCounts.singleValueColumnCount
                );
            }
        );
    });
}

/*------------------------------------------------------------------------------
 * Robot._getCategoryCount()
 * 
 * Calculates the number of categories visited. Here a category is defined as
 * a unique combination of navigation variables.
 *------------------------------------------------------------------------------
 */
Robot.prototype._getCategoryCount = function() {
    
    var uniqCats = [];
    var todoList = this._todoList;
    var pos = todoList.getFirst();
    
    while(pos) {
        var navVars = todoList.getData(pos).getVariables();
        var varNames = Object.keys(navVars);
        
        for (var i=0; i<uniqCats.length; i++) {
            for (var j = 0; j<varNames.length; j++) {
                if (uniqCats[i][varNames[j]] != navVars[varNames[j]]) {
                    break;
                }
            }
            if (j == varNames.length) { // Category found
                break;
            }
        }
        if (i == uniqCats.length) {
            uniqCats.push(navVars);
        }
        pos = todoList.getNext(pos);
    }
    
    return uniqCats.length;
/*
    var record = { count: uniqCats.length };
    
    var newFile = csv.format(EXCEL_CSV_NL);
    newFile.pipe(fs.createWriteStream(this._catFilename));
    newFile.write(record);
*/
}

/*------------------------------------------------------------------------------
 * Robot._saveCounts()
 * 
 * Saves all the counts generated by this robot run to the counts file.
 *------------------------------------------------------------------------------
 */
Robot.prototype._saveCounts = function(catCount, emptyCount, singleCount) {
    
    var record = {
        categoryCount:          catCount,
        emptyColumnCount:       emptyCount,
        singleValueColumnCount: singleCount
    };
    
    var newFile = csv.format(EXCEL_CSV_NL);
    newFile.pipe(fs.createWriteStream(this._countsFilename));
    newFile.write(record);
}

module.exports = Robot;
